//////////////////////////////////////////////////////////
// Main CPU routines

#ifdef MSS
#include <mss.h>
#endif

#include <stdio.h>
#include <stdlib.h> 
#include <string.h>
#include <ctype.h>

#include "vb_types.h"
#include "v810_ins.h"
#include "v810_opt.h"
#include "v810_cpu.h"
#include "vb_vbt.h" //Remove Me!!!


////////////////////////////////////////////////////////////
// Globals
WORD P_REG[32];  // Main program reg pr0-pr31
WORD S_REG[32];  // System registers sr0-sr31
WORD PC;         // Program Counter


// Reinitialize the defaults in the CPU
void v810_reset() {
    P_REG[0]      =  0x00000000;
    PC            =  0xFFFFFFF0;
    S_REG[ECR]    =  0x0000FFF0;
    S_REG[PSW]    =  0x00008000;
    S_REG[PIR]    =  0x00008300;
    S_REG[TKCW]   =  0x000000E0;
    //S_REG[HCCW] =  0x0000FFF0; //V830?
}

long v810_dis(long tPC, int num) {
    int      lowB, highB, lowB2, highB2;             // up to 4 bytes for instruction (either 16 or 32 bits)
    int      opcode, arg1, arg2, arg3;
    int i = 0;

    //if (tPC == -1) // | tPC < V810_ROM1.lowaddr
    //      tPC = PC;
    for(i = 0; i< num; i++) {
        lowB   = mem_rbyte(tPC);
        highB  = mem_rbyte(tPC+1);
        lowB2  = mem_rbyte(tPC+2);
        highB2 = mem_rbyte(tPC+3);
    
        opcode = highB >> 2;
        if((highB & 0xE0) == 0x80)        // Special opcode format for          
            opcode = (highB >> 1);            // type III instructions.
    
        if((opcode > 0x4F) | (opcode < 0)) {
            //Error Invalid opcode!
            printf("\n%08x\t\t%2x %2x  ;Invalid Opcode", tPC, lowB, highB);
            tPC += 2;                                               
        }
        
        switch(optable[opcode].addr_mode) {
        case AM_I:       // Do the same Ither way =)
            arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
            arg2 = (lowB & 0x1F);
            if (opcode == JMP) {
                printf("\n%08x\t%s\t\t[$%d]", tPC, optable[opcode].opname, arg2);
            } else {
                printf("\n%08x\t%s\t\t$%d, $%d", tPC, optable[opcode].opname, arg2, arg1);
            }
            tPC += 2;   // 16 bit instruction
            break;
        case AM_II:
            arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
            arg2 = (lowB & 0x1F);
            printf("\n%08x\t%s\t\t%d, $%d", tPC, optable[opcode].opname, sign_5(arg2), arg1);
            tPC += 2;   // 16 bit instruction
            break;
        case AM_III:
            arg1 = ((highB & 0x1) << 8) + (lowB & 0xFE);
            printf("\n%08x\t%s\t\t%d", tPC, optable[opcode].opname, sign_9(arg1));
            tPC += 2;   // 16 bit instruction
            break;
        case AM_IV:
            arg1 = ((highB & 0x3) << 24) + (lowB << 16) + (highB2 << 8) + lowB2;
            printf("\n%08x\t%s\t\t%d", tPC, optable[opcode].opname, sign_26(arg1));
            tPC += 4;                                               // 32 bit instruction
            break;
        case AM_V:       
            arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
            arg2 = (lowB & 0x1F);
            arg3 = (highB2 << 8) + lowB2;
            printf("\n%08x\t%s\t\t%x, $%d, $%d", tPC, optable[opcode].opname, arg3, arg2, arg1 );
            tPC += 4;   // 32 bit instruction
            break;
        case AM_VIa:  // Mode6 form1
            arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
            arg2 = (lowB & 0x1F);
            arg3 = (highB2 << 8) + lowB2;
            printf("\n%08x\t%s\t\t%d[$%d], $%d", tPC, optable[opcode].opname, sign_16(arg3), arg2, arg1);
            tPC += 4;   // 32 bit instruction
            break;
        case AM_VIb:  // Mode6 form2
            arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
            arg2 = (lowB & 0x1F);
            arg3 = (highB2 << 8) + lowB2;                              //  whats the order??? 2,3,1 or 1,3,2
            printf("\n%08x\t%s\t\t$%d, %d[$%d]", tPC, optable[opcode].opname, arg1, sign_16(arg3), arg2);
            tPC += 4;   // 32 bit instruction
            break;
        case AM_VII:   // Unhandled
            printf("\n%08x\t\t%2x %2x %2x %2x", tPC, lowB, highB, lowB2, highB2);
            tPC +=4;        // 32 bit instruction
            break;
        case AM_VIII:  // Unhandled
            printf("\n%08x\t\t%2x %2x %2x %2x", tPC, lowB, highB, lowB2, highB2);
            tPC += 4;   // 32 bit instruction
            break;
        case AM_IX:
            arg1 = (lowB & 0x1); // Mode ID, Ignore for now
            printf("\n%08x\t%s", tPC, optable[opcode].opname);   
            tPC += 2;   // 16 bit instruction
            break;
        case AM_BSTR:  // Bit String Subopcodes
            arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
            arg2 = (lowB & 0x1F);
            if(arg2 > 16) {
                printf("\n%08x\tBError", tPC);
            } else {
                printf("\n%08x\t%s, $%d", PC, bssuboptable[arg2].opname,arg1);
            }
            tPC += 2;   // 16 bit instruction
            break;
        case AM_FPP:   // Floating Point Subcode
            arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
            arg2 = (lowB & 0x1F);
            arg3 = (highB2 >> 2);
            if(arg3 > 16) {
                printf("\n%08x\tFError", tPC);
            } else {
                printf("\n%08x\t%s  $%d, $%d", PC, fpsuboptable[arg3].opname, arg1, arg2);
            }
            tPC += 4;   // 32 bit instruction
            break;
        case AM_UDEF:  // Invalid opcode.
            default:           // Invalid opcode.
            printf("\n%08x\t\t%2x %2x  ;Invalid Opcode", tPC, lowB, highB);
            tPC += 2;                                               
        }
    }
    return(tPC);
}

void v810_preg() {
    int i = 0;
    //int j = 0;
    for(i = 0; i < 32; i+=4) {
        //for(j = 0; j < 4; j++) {
        printf("\n$%02d %08x, $%02d %08x, $%02d %08x, $%02d %08x",i, P_REG[i],i+1,P_REG[i+1],i+2,P_REG[i+2],i+3,P_REG[i+3]);
    } 
    printf("\n$PSW %08x",S_REG[PSW]);
}

void v810_dump(long tPC, int num) {
    BYTE t1, t2, t3, t4;
    int i = 0;
    int j = 0;
    
    //if (tPC == -1) // | tPC < V810_ROM1.lowaddr
    //      tPC = PC;
    
    for(i = 0; i< num; i++) {
        printf("\n%08x",tPC);
        for (j = 0; j < 16; j+=8) {
            t1 = mem_rhword(tPC+j);
            t2 = mem_rhword(tPC+j+2);
            t3 = mem_rhword(tPC+j+4);
            t4 = mem_rhword(tPC+j+6);
            printf("  %04x %04x %04x%04x", t1, t2, t3, t4);
        }
        tPC +=16;
    }
}

//could be done a lot better, later maby!
void v810_trc(int num, int noKey) {
    int      lowB, highB, lowB2, highB2;             // up to 4 bytes for instruction (either 16 or 32 bits)
    int      opcode, arg1, arg2, arg3;
    int Ttmp;
    //WORD tmpWord;
    HWORD tmpHWordL;
    HWORD tmpHWordH;
    int i;
    int step;

    for(i = 0; i< num; i++) {
        if(!noKey) {
            if(kbhit())
                if(getch() == 27) break; //Fix with Allegro!
        }

        //Adjust the Timer.... should not be hear!
        //Many bugs, look into it more...
        if(tHReg.TCR & 0x01) { // Timer Enabled
		step = 1;
		if(tHReg.TCR & 0x10) step = 5;

            //if(tHReg.tTHW >= step) {
            //    tHReg.tTHW -=step;

            //Decrement Timer.... (old)
            Ttmp = tHReg.TLB + (tHReg.THB << 8);
            Ttmp-=step;
            tHReg.TLB = (Ttmp&0xFF);
            tHReg.THB = ((Ttmp >> 8)&0xFF);

            if(Ttmp < step) {
            //} else { //Ran Down to zero
                //Reset the internal timmer
                //tHReg.tTHW  = (((HWORD)tHReg.TLB)|(((HWORD)tHReg.THB) << 8));
                tHReg.TCR = tHReg.TCR | 0x02; //Zero Status
                if(tHReg.TCR & 0x08) { //Timer Interupt Enabled
                    //if(tVIPREG.INTENB&0xFFFF) { //Wrong Number???
                        v810_int(1); 
                    //}
                }
            }
        }

        //Faster? change the code to use hWords....
        tmpHWordL = mem_rhword(PC);
        tmpHWordH = mem_rhword(PC+2);
        lowB = ((tmpHWordL)&0xFF);
        highB = ((tmpHWordL>>8)&0xFF);
        lowB2 = ((tmpHWordH)&0xFF);
        highB2 = ((tmpHWordH>>8)&0xFF);

//This gives me errors, Why!!!! (Dont always get the same data
//        tmpWord = mem_rword(PC);
//        if(lowB != ((tmpWord)&0xFF)) printf("Low %d %d",lowB,(tmpWord)&0xFF);
//        if(highB != ((tmpWord>>8)&0xFF)) printf("hi %d %d",highB,(tmpWord>>8)&0xFF);
//        if(lowB2 != ((tmpWord>>16)&0xFF)) printf("Low2 %d% d",lowB2,(tmpWord>>16)&0xFF);
//        if(highB2 != ((tmpWord>>24)&0xFF)) printf("hi2 %d %d",highB2,(tmpWord>>24)&0xFF);

//Remove Me!!!
P_REG[0]=0; //Zero the Zero Reg!!!

        opcode = highB >> 2;
        if((highB & 0xE0) == 0x80)        // Special opcode format for          
            opcode = (highB >> 1);    // type III instructions.

        if((opcode > 0x4F) | (opcode < 0)) {
            //Error Invalid opcode!
            //fprintf(ferr,"\n%08x\t\t%2x %2x  ;Invalid Opcode", PC, lowB, highB);
            PC += 2;                                                
        }
        
        switch(optable[opcode].addr_mode) {
          case AM_I:       // Do the same Ither way =)
            arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
            arg2 = (lowB & 0x1F);
            PC += 2;   // 16 bit instruction
            if (opcode == JMP) {
                // Make a call to our fnc, by derefrencing the addres from our array
                (*optable[opcode].func)(arg2, 0, 0);       
            } else {
                (*optable[opcode].func)(arg2, arg1, 0);    
            }
            break;
          case AM_II:
            arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
            arg2 = (lowB & 0x1F);
            PC += 2;   // 16 bit instruction
            (*optable[opcode].func)(arg2, arg1, 0);    
            break;
          case AM_III:
            arg1 = ((highB & 0x1) << 8) + (lowB & 0xFE);
            (*optable[opcode].func)(arg1, 0, 0);       
            // let inst inc pc, dont always need to you know!
            break;
          case AM_IV:
            arg1 = ((highB & 0x3) << 24) + (lowB << 16) + (highB2 << 8) + lowB2;
            //PC += 4;  //only jumps hear                           // 32 bit instruction
            (*optable[opcode].func)(arg1, 0, 0);       
            break;
          case AM_V:       
            arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
            arg2 = (lowB & 0x1F);
            arg3 = (highB2 << 8) + lowB2;
            PC += 4;   // 32 bit instruction
            (*optable[opcode].func)(arg3, arg2, arg1);         
            break;
          case AM_VIa:  // Mode6 form1
            arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
            arg2 = (lowB & 0x1F);
            arg3 = (highB2 << 8) + lowB2;
            PC += 4;   // 32 bit instruction
            (*optable[opcode].func)(arg3, arg2, arg1);         
            break;
          case AM_VIb:  // Mode6 form2
            arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
            arg2 = (lowB & 0x1F);
            arg3 = (highB2 << 8) + lowB2;                              //  whats the order??? 2,3,1 or 1,3,2
            PC += 4;   // 32 bit instruction
            (*optable[opcode].func)(arg1, arg3, arg2);         
            break;
          case AM_VII:   // Unhandled
            fprintf(ferr,"\n%08x\t\t%2x %2x %2x %2x", PC, lowB, highB, lowB2, highB2);
            PC +=4; // 32 bit instruction
            break;
          case AM_VIII:  // Unhandled
            fprintf(ferr,"\n%08x\t%\t2x %2x %2x %2x", PC, lowB, highB, lowB2, highB2);
            PC += 4;   // 32 bit instruction
            break;
          case AM_IX:
            arg1 = (lowB & 0x1); // Mode ID, Ignore for now
            PC += 2;   // 16 bit instruction
            (*optable[opcode].func)(arg1, 0, 0);       
            break;
          case AM_BSTR:  // Bit String Subopcodes
            arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
            arg2 = (lowB & 0x1F);
            PC += 2;   // 16 bit instruction
            if(arg2 > 16) {
                fprintf(ferr,"\n%08x\tBSR Error: %04x", PC,arg2);
            } else {
                (*optable[opcode].func)(arg1, arg2, 0);    
                //fprintf(ferr,"\n%08x\t%s, $%d", PC, bssuboptable[arg2].opname,arg1);
            }
            break;
          case AM_FPP:   // Floating Point Subcode
            arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
            arg2 = (lowB & 0x1F);
            arg3 = ((highB2 >> 2)&0x3F);
            PC += 4;   // 32 bit instruction
            if(arg3 > 16) {
                fprintf(ferr,"\n%08x\tFPU Error: %04x", PC, arg3);
            } else {
                (*optable[opcode].func)(arg1, arg2, arg3);         
                //fprintf(ferr,"\n%08x\t%s  $%d, $%d", PC, fpsuboptable[arg3].opname, arg1, arg2);
            }
            break;
          case AM_UDEF:  // Invalid opcode.
          default:           // Invalid opcode.
            fprintf(ferr,"\n%08x\t\t%2x %2x  ;Invalid Opcode", PC, lowB, highB);
            PC += 2;                                                
        }
    }
}

// Generate Interupt #n
void v810_int(WORD iNum) {
    if (iNum > 0x0F) return;  // Invalid Interupt number...
    if((S_REG[PSW] & PSW_NP)) return;
    if((S_REG[PSW] & PSW_EP)) return;
    if((S_REG[PSW] & PSW_ID)) return;

    if(iNum < ((S_REG[PSW] & PSW_IA)>>16)) return; // Interupt to low on the chain

    //WORD tData = mem_rword(0xFFFFFE00 | (iNum << 4));
    //if((tData  == 0xFFFF)|(tData == 0x0000)|(tData == 0x008A)) return; // Dont jump to Invalid Interupts...
    //if(!tVIPREG.INTENB) return;

    //fprintf(ferr,"Int %x\n",iNum);
    //Ready to Generate the Interupts
    S_REG[EIPC] = PC;
    S_REG[EIPSW] = S_REG[PSW];

    PC = 0xFFFFFE00 | (iNum << 4);
    
    S_REG[ECR] = 0xFE00 | (iNum << 8);
    S_REG[PSW] = S_REG[PSW] | PSW_EP;
    S_REG[PSW] = S_REG[PSW] | PSW_ID;
    if((iNum+=1) > 0x0F) (iNum = 0x0F);
    S_REG[PSW] = S_REG[PSW] | (iNum << 16); //Set the Interupt

    tVIPREG.INTPND = tVIPREG.INTENB; //Need to decode the actual Int some day... (how?)
}

/*
// Generate exception #n
//Exceptions are Div by zero, trap and Invalid Opcode, we can live without...
void v810_exp(int iNum) {
    if (iNum > 0x0F) return;  // Invalid Interupt number...
    if(!S_REG[PSW]&PSW_ID) return; 

    if(iNum < ((S_REG[PSW] & PSW_IA)>>16)) return; // Interupt to low on the mask level....
    if(S_REG[PSW]&PSW_NP) { //Fatal Exception
        S_REG[DPC] = PC;
        S_REG[DPSW] = S_REG[PSW];
        //ECR.EICC = Exception Code, dont get it???
        S_REG[PSW] = S_REG[PSW] | PSW_DP; 
        S_REG[PSW] = S_REG[PSW] | PSW_NP; 
        S_REG[PSW] = S_REG[PSW] | PSW_EP; 
        S_REG[PSW] = S_REG[PSW] | PSW_ID; 
        S_REG[PSW] = S_REG[PSW] | (((iNum+1) & 0x0f) << 16); //Set the Interupt status
        
        PC = 0xFFFFFFE0;
        return;
    }else if(S_REG[PSW]&PSW_EP) { //Double Exception
        S_REG[FEPC] = PC;
        S_REG[FEPSW] = S_REG[PSW];
        //ECR.FECC = Exception Code, dont get it???
        S_REG[PSW] = S_REG[PSW] | PSW_NP; 
        S_REG[PSW] = S_REG[PSW] | PSW_ID; 
        S_REG[PSW] = S_REG[PSW] | (((iNum+1) & 0x0f) << 16); //Set the Interupt status
    
        PC = 0xFFFFFFD0;
        return;
    } else {                                // Regular Exception
        S_REG[EIPC] = PC;
        S_REG[EIPSW] = S_REG[PSW];
        //ECR.EICC = Exception Code, dont get it???
        S_REG[PSW] = S_REG[PSW] | PSW_EP; 
        S_REG[PSW] = S_REG[PSW] | PSW_ID; 
        S_REG[PSW] = S_REG[PSW] | (((iNum+1) & 0x0f) << 16); //Set the Interupt status
        return;
    }
}
*/

